/*
 * lib/cachectl.S
 *
 * Copyright (C) 1994, 1995 by Waldorf Electronics,
 * written by Ralf Baechle
 *
 * This file is subject to the terms and conditions of the
 * GNU General Public License. See the file COPYING in the
 * main directory of this archive for more details.
 */
#include <asm/asm.h>
#include <asm/cachectl.h>
#include <asm/mipsregs.h>
#include <asm/segment.h>

/*
 * Some bits in the config register
 */
#define CONFIG_IB       (1<<5)
#define CONFIG_DB       (1<<4)

/*
 * Flush instruction/data caches
 *
 * Parameters: a0 - starting address to flush
 *             a1 - size of area to be flushed
 *             a2 - which caches to be flushed
 *
 * FIXME:      - ignores parameters in a0/a1
 *             - doesn't know about second level caches
 */

		.text

		.set	noreorder
		LEAF(cacheflush)
		andi	t1,a2,DCACHE
		beqz	t1,do_icache
		li	t0,KSEG0		# delay slot

		/*
		 * Writeback data cache, even lines
		 */
		li	t1,CACHELINES-1
1:		cache	1,0(t0)
		cache	1,32(t0)
		cache	1,64(t0)
		cache	1,96(t0)
		cache	1,128(t0)
		cache	1,160(t0)
		cache	1,192(t0)
		cache	1,224(t0)
		cache	1,256(t0)
		cache	1,288(t0)
		cache	1,320(t0)
		cache	1,352(t0)
		cache	1,384(t0)
		cache	1,416(t0)
		cache	1,448(t0)
		cache	1,480(t0)
		addiu	t0,512
		bnez	t1,1b
		subu	t1,1

		/*
		 * Writeback data cache, odd lines
		 * Only needed for 16 byte line size
		 */
		mfc0	t1,CP0_CONFIG
		andi	t1,CONFIG_IB
		bnez	t1,do_icache
		li	t1,CACHELINES-1
1:		cache	1,16(t0)
		cache	1,48(t0)
		cache	1,80(t0)
		cache	1,112(t0)
		cache	1,144(t0)
		cache	1,176(t0)
		cache	1,208(t0)
		cache	1,240(t0)
		cache	1,272(t0)
		cache	1,304(t0)
		cache	1,336(t0)
		cache	1,368(t0)
		cache	1,400(t0)
		cache	1,432(t0)
		cache	1,464(t0)
		cache	1,496(t0)
		addiu	t0,512
		bnez	t1,1b
		subu	t1,1

do_icache:	andi	t1,a2,ICACHE
		beqz	t1,done

		/*
		 * Flush instruction cache, even lines
		 */
		lui	t0,0x8000
		li	t1,CACHELINES-1
1: 		cache	0,0(t0)	
		cache	0,32(t0)
		cache	0,64(t0)
		cache	0,96(t0)
		cache	0,128(t0)
		cache	0,160(t0)
		cache	0,192(t0)
		cache	0,224(t0)
		cache	0,256(t0)
		cache	0,288(t0)
		cache	0,320(t0)
		cache	0,352(t0)
		cache	0,384(t0)
		cache	0,416(t0)
		cache	0,448(t0)
		cache	0,480(t0)
		addiu	t0,512
		bnez	t1,1b
		subu	t1,1

		/*
		 * Flush instruction cache, even lines
		 * Only needed for 16 byte line size
		 */
		mfc0	t1,CP0_CONFIG
		andi	t1,CONFIG_DB
		bnez	t1,done
		li	t1,CACHELINES-1
1:		cache	0,16(t0)
		cache	0,48(t0)
		cache	0,80(t0)
		cache	0,112(t0)
		cache	0,144(t0)
		cache	0,176(t0)
		cache	0,208(t0)
		cache	0,240(t0)
		cache	0,272(t0)
		cache	0,304(t0)
		cache	0,336(t0)
		cache	0,368(t0)
		cache	0,400(t0)
		cache	0,432(t0)
		cache	0,464(t0)
		cache	0,496(t0)
		addiu	t0,512
		bnez	t1,1b
		subu	t1,1

done:		j	ra
		nop
		END(cacheflush)
